from django.conf.urls import url

from appalerts import views, views_ajax

app_name = 'appalerts'

urlpatterns = [
    url(
        regex=r'^list$',
        view=views.AppAlertListView.as_view(),
        name='list',
    ),
    url(
        regex=r'^redirect/(?P<appalert>[0-9a-f-]+)$',
        view=views.AppAlertLinkRedirectView.as_view(),
        name='redirect',
    ),
    url(
        regex=r'^ajax/status$',
        view=views_ajax.AppAlertStatusAjaxView.as_view(),
        name='ajax_status',
    ),
    url(
        regex=r'^ajax/dismiss/(?P<appalert>[0-9a-f-]+)$',
        view=views_ajax.AppAlertDismissAjaxView.as_view(),
        name='ajax_dismiss',
    ),
    url(
        regex=r'^ajax/dismiss/all$',
        view=views_ajax.AppAlertDismissAjaxView.as_view(),
        name='ajax_dismiss_all',
    ),
]



"""Taskflow tests for management commands in the projectroles app"""

# NOTE: You must supply 'sodar_url': self.live_server_url in taskflow requests!
#       This is due to the Django 1.10.x feature described here:
#       https://code.djangoproject.com/ticket/27596

from django.conf import settings
from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from django.forms.models import model_to_dict
from django.test import (
    LiveServerTestCase,
    RequestFactory,
    tag,
    override_settings,
)

# HACK to get around https://stackoverflow.com/a/25081791
from django.urls import reverse

from unittest import skipIf

from projectroles import views_taskflow
from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    Project,
    Role,
    RoleAssignment,
    ProjectInvite,
    SODAR_CONSTANTS,
)
from projectroles.plugins import get_backend_api, change_plugin_status
from projectroles.tests.taskflow_testcase import TestCase
from projectroles.tests.test_models import (
    ProjectInviteMixin,
    ProjectMixin,
    RoleAssignmentMixin,
)
from projectroles.tests.test_views import TestViewsBase, TASKFLOW_SECRET_INVALID

# Access Django user model
User = auth.get_user_model()

# App settings API
app_settings = AppSettingAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SUBMIT_STATUS_OK = SODAR_CONSTANTS['SUBMIT_STATUS_OK']
SUBMIT_STATUS_PENDING = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
SUBMIT_STATUS_PENDING_TASKFLOW = SODAR_CONSTANTS[
    'SUBMIT_STATUS_PENDING_TASKFLOW'
]
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']

# Local constants
INVITE_EMAIL = 'test@example.com'
SECRET = 'rsd886hi8276nypuvw066sbvv0rb2a6x'
TASKFLOW_ENABLED = (
    True if 'taskflow' in settings.ENABLED_BACKEND_PLUGINS else False
)
TASKFLOW_SKIP_MSG = 'Taskflow not enabled in settings'
TASKFLOW_TEST_MODE = getattr(settings, 'TASKFLOW_TEST_MODE', False)


# Base Classes -----------------------------------------------------------------


@tag('Taskflow')  # Run tests if iRODS and SODAR Taskflow are enabled
class TestTaskflowBase(
    ProjectMixin, RoleAssignmentMixin, LiveServerTestCase, TestCase
):
    """Base class for testing UI views with taskflow"""

    def _make_project_taskflow(
        self, title, type, parent, owner, description, public_guest_access=False
    ):
        """Make Project with taskflow for UI view tests"""
        post_data = dict(self.request_data)
        post_data.update(
            {
                'title': title,
                'type': type,
                'parent': parent.sodar_uuid if parent else None,
                'owner': owner.sodar_uuid,
                'description': description,
                'public_guest_access': public_guest_access,
            }
        )
        post_data.update(
            app_settings.get_all_defaults(
                APP_SETTING_SCOPE_PROJECT, post_safe=True
            )
        )  # Add default settings

        post_kwargs = {'project': parent.sodar_uuid} if parent else {}

        with self.login(self.user):
            response = self.client.post(
                reverse('projectroles:create', kwargs=post_kwargs), post_data
            )
            self.assertEqual(response.status_code, 302)
            project = Project.objects.get(title=title)
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:detail',
                    kwargs={'project': project.sodar_uuid},
                ),
            )

        owner_as = project.get_owner()
        return project, owner_as

    def _make_assignment_taskflow(self, project, user, role):
        """Make RoleAssignment with taskflow for UI view tests"""
        post_data = dict(self.request_data)
        post_data.update(
            {
                'project': project.sodar_uuid,
                'user': user.sodar_uuid,
                'role': role.pk,
            }
        )

        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:role_create',
                    kwargs={'project': project.sodar_uuid},
                ),
                post_data,
            )
            role_as = RoleAssignment.objects.get(project=project, user=user)
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:roles', kwargs={'project': project.sodar_uuid}
                ),
            )

        return role_as

    def setUp(self):
        # Ensure TASKFLOW_TEST_MODE is True to avoid data loss
        if not TASKFLOW_TEST_MODE:
            raise ImproperlyConfigured(
                'TASKFLOW_TEST_MODE not True, '
                'testing with SODAR Taskflow disabled'
            )

        # Set up live server URL for requests
        self.request_data = {'sodar_url': self.live_server_url}

        # Get taskflow plugin (or None if taskflow not enabled)
        change_plugin_status(
            name='taskflow', status=0, plugin_type='backend'  # 0 = Enabled
        )
        self.taskflow = get_backend_api('taskflow', force=True)

        # Init roles
        self.role_owner = Role.objects.get_or_create(name=PROJECT_ROLE_OWNER)[0]
        self.role_delegate = Role.objects.get_or_create(
            name=PROJECT_ROLE_DELEGATE
        )[0]
        self.role_contributor = Role.objects.get_or_create(
            name=PROJECT_ROLE_CONTRIBUTOR
        )[0]
        self.role_guest = Role.objects.get_or_create(name=PROJECT_ROLE_GUEST)[0]

        # Init users
        self.user_cat = self.make_user('user_cat')
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()

        # Create category locally
        self.category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(self.category, self.user_cat, self.role_owner)

    def tearDown(self):
        self.taskflow.cleanup()


# Local Tests ------------------------------------------------------------------


class TestTaskflowLocalAPIBase(
    ProjectMixin, RoleAssignmentMixin, TestViewsBase
):
    """Base class for testing the local taskflow API views"""

    def setUp(self):
        super().setUp()
        self.req_factory = RequestFactory()


@override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
class TestProjectGetAPIView(TestTaskflowLocalAPIBase):
    """Tests for the project retrieve API view"""

    def setUp(self):
        super().setUp()
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post(self):
        """Test POST request for getting a project"""
        request = self.req_factory.post(
            reverse('projectroles:taskflow_project_get'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowProjectGetAPIView.as_view()(request)
        self.assertEqual(response.status_code, 200)
        expected = {
            'project_uuid': str(self.project.sodar_uuid),
            'title': self.project.title,
            'description': self.project.description,
        }
        self.assertEqual(response.data, expected)

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_get_pending(self):
        """Test POST request to get a pending project"""
        pd_project = self._make_project(
            title='TestProject2',
            type=PROJECT_TYPE_PROJECT,
            parent=None,
            submit_status=SUBMIT_STATUS_PENDING_TASKFLOW,
        )
        request = self.req_factory.post(
            reverse('projectroles:taskflow_project_get'),
            data={
                'project_uuid': str(pd_project.sodar_uuid),
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowProjectGetAPIView.as_view()(request)
        self.assertEqual(response.status_code, 404)


@override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
class TestProjectUpdateAPIView(TestTaskflowLocalAPIBase):
    """Tests for the project updating API view"""

    def setUp(self):
        super().setUp()
        self.category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self._make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post(self):
        """Test POST request for updating a project"""
        # NOTE: Duplicate titles not checked here, not allowed in the form
        title = 'New title'
        desc = 'New desc'
        readme = 'New readme'
        request = self.req_factory.post(
            reverse('projectroles:taskflow_project_update'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'title': title,
                'parent_uuid': str(self.category.sodar_uuid),
                'description': desc,
                'readme': readme,
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowProjectUpdateAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.project.refresh_from_db()
        self.assertEqual(self.project.title, title)
        self.assertEqual(self.project.description, desc)
        self.assertEqual(self.project.readme.raw, readme)

    def test_post_no_description(self):
        """Test POST request without a description field"""
        # NOTE: Duplicate titles not checked here, not allowed in the form
        title = 'New title'
        readme = 'New readme'
        request = self.req_factory.post(
            reverse('projectroles:taskflow_project_update'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'title': title,
                'parent_uuid': str(self.category.sodar_uuid),
                'readme': readme,
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowProjectUpdateAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.project.refresh_from_db()
        self.assertEqual(self.project.title, title)
        self.assertEqual(self.project.description, '')
        self.assertEqual(self.project.readme.raw, readme)

    def test_post_move(self):
        """Test POST request for moving a project"""
        new_category = self._make_project('NewCat', PROJECT_TYPE_CATEGORY, None)
        request = self.req_factory.post(
            reverse('projectroles:taskflow_project_update'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'title': self.project.title,
                'parent_uuid': str(new_category.sodar_uuid),
                'description': self.project.description,
                'readme': '',
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowProjectUpdateAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.project.refresh_from_db()
        self.assertEqual(self.project.parent, new_category)


@override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
class TestRoleAssignmentGetAPIView(TestTaskflowLocalAPIBase):
    """Tests for the role assignment getting API view"""

    def setUp(self):
        super().setUp()
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post(self):
        """Test POST request for getting a role assignment"""
        request = self.req_factory.post(
            reverse('projectroles:taskflow_role_get'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'user_uuid': str(self.user.sodar_uuid),
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowRoleAssignmentGetAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        expected = {
            'assignment_uuid': str(self.owner_as.sodar_uuid),
            'project_uuid': str(self.project.sodar_uuid),
            'user_uuid': str(self.user.sodar_uuid),
            'role_pk': self.role_owner.pk,
            'role_name': self.role_owner.name,
        }
        self.assertEqual(response.data, expected)


@override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
class TestRoleAssignmentSetAPIView(TestTaskflowLocalAPIBase):
    """Tests for the role assignment setting API view"""

    def setUp(self):
        super().setUp()
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post_new(self):
        """Test POST request for assigning a new role"""
        new_user = self.make_user('new_user')
        self.assertEqual(RoleAssignment.objects.all().count(), 1)
        request = self.req_factory.post(
            reverse('projectroles:taskflow_role_set'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'user_uuid': str(new_user.sodar_uuid),
                'role_pk': self.role_contributor.pk,
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowRoleAssignmentSetAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(RoleAssignment.objects.all().count(), 2)
        new_as = RoleAssignment.objects.get(project=self.project, user=new_user)
        self.assertEqual(new_as.role.pk, self.role_contributor.pk)

    def test_post_existing(self):
        """Test POST request for updating an existing role"""
        new_user = self.make_user('new_user')
        self._make_assignment(self.project, new_user, self.role_guest)
        self.assertEqual(RoleAssignment.objects.all().count(), 2)
        request = self.req_factory.post(
            reverse('projectroles:taskflow_role_set'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'user_uuid': str(new_user.sodar_uuid),
                'role_pk': self.role_contributor.pk,
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowRoleAssignmentSetAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(RoleAssignment.objects.all().count(), 2)

        new_as = RoleAssignment.objects.get(project=self.project, user=new_user)
        self.assertEqual(new_as.role.pk, self.role_contributor.pk)


@override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
class TestRoleAssignmentDeleteAPIView(TestTaskflowLocalAPIBase):
    """Tests for the role assignment deletion API view"""

    def setUp(self):
        super().setUp()
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post(self):
        """Test POST request for removing a role assignment"""
        new_user = self.make_user('new_user')
        self._make_assignment(self.project, new_user, self.role_guest)
        self.assertEqual(RoleAssignment.objects.all().count(), 2)
        request = self.req_factory.post(
            reverse('projectroles:taskflow_role_delete'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'user_uuid': str(new_user.sodar_uuid),
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowRoleAssignmentDeleteAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(RoleAssignment.objects.all().count(), 1)

    def test_post_not_found(self):
        """Test POST request for removing a non-existing role assignment"""
        new_user = self.make_user('new_user')
        self.assertEqual(RoleAssignment.objects.all().count(), 1)
        request = self.req_factory.post(
            reverse('projectroles:taskflow_role_delete'),
            data={
                'project_uuid': str(self.project.sodar_uuid),
                'user_uuid': str(new_user.sodar_uuid),
                'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
            },
        )
        response = views_taskflow.TaskflowRoleAssignmentDeleteAPIView.as_view()(
            request
        )
        self.assertEqual(response.status_code, 404)
        self.assertEqual(RoleAssignment.objects.all().count(), 1)


class TestTaskflowAPIViewAccess(TestTaskflowLocalAPIBase):
    """Tests for taskflow API view access"""

    def setUp(self):
        super().setUp()
        self.category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_access_invalid_token(self):
        """Test access with an invalid token"""
        urls = [
            reverse('projectroles:taskflow_project_get'),
            reverse('projectroles:taskflow_project_update'),
            reverse('projectroles:taskflow_role_get'),
            reverse('projectroles:taskflow_role_set'),
            reverse('projectroles:taskflow_role_delete'),
            reverse('projectroles:taskflow_settings_get'),
            reverse('projectroles:taskflow_settings_set'),
        ]
        for url in urls:
            request = self.req_factory.post(
                url, data={'sodar_secret': TASKFLOW_SECRET_INVALID}
            )
            response = views_taskflow.TaskflowProjectGetAPIView.as_view()(
                request
            )
            self.assertEqual(response.status_code, 403)

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_access_no_token(self):
        """Test access with no token"""
        urls = [
            reverse('projectroles:taskflow_project_get'),
            reverse('projectroles:taskflow_project_update'),
            reverse('projectroles:taskflow_role_get'),
            reverse('projectroles:taskflow_role_set'),
            reverse('projectroles:taskflow_role_delete'),
            reverse('projectroles:taskflow_settings_get'),
            reverse('projectroles:taskflow_settings_set'),
        ]
        for url in urls:
            request = self.req_factory.post(url)
            response = views_taskflow.TaskflowProjectGetAPIView.as_view()(
                request
            )
            self.assertEqual(response.status_code, 403)

    @override_settings(ENABLED_BACKEND_PLUGINS=[])
    def test_disable_api_views(self):
        """Test to make sure API views are disabled without taskflow"""
        urls = [
            reverse('projectroles:taskflow_project_get'),
            reverse('projectroles:taskflow_project_update'),
            reverse('projectroles:taskflow_role_get'),
            reverse('projectroles:taskflow_role_set'),
            reverse('projectroles:taskflow_role_delete'),
            reverse('projectroles:taskflow_settings_get'),
            reverse('projectroles:taskflow_settings_set'),
        ]
        for url in urls:
            request = self.req_factory.post(
                url, data={'sodar_secret': settings.TASKFLOW_SODAR_SECRET}
            )
            response = views_taskflow.TaskflowProjectGetAPIView.as_view()(
                request
            )
            self.assertEqual(response.status_code, 403)


# Tests Requiring iRODS --------------------------------------------------------


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestProjectCreateView(TestTaskflowBase):
    """Tests for Project creation view with taskflow"""

    def test_create_project(self):
        """Test Project creation with taskflow"""

        # Assert precondition
        self.assertEqual(Project.objects.all().count(), 1)

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Assert Project state after creation
        self.assertEqual(Project.objects.all().count(), 2)
        project = Project.objects.all()[0]
        self.assertIsNotNone(project)

        expected = {
            'id': project.pk,
            'title': 'TestProject',
            'type': PROJECT_TYPE_PROJECT,
            'parent': self.category.pk,
            'submit_status': SUBMIT_STATUS_OK,
            'description': 'description',
            'public_guest_access': False,
            'full_title': self.category.title + ' / TestProject',
            'has_public_children': False,
            'sodar_uuid': project.sodar_uuid,
        }

        model_dict = model_to_dict(project)
        model_dict.pop('readme', None)
        self.assertEqual(model_dict, expected)

        # Assert owner role assignment
        owner_as = RoleAssignment.objects.get(
            project=project, role=self.role_owner
        )

        expected = {
            'id': owner_as.pk,
            'project': project.pk,
            'role': self.role_owner.pk,
            'user': self.user.pk,
            'sodar_uuid': owner_as.sodar_uuid,
        }

        self.assertEqual(model_to_dict(owner_as), expected)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestProjectUpdateView(TestTaskflowBase):
    """Tests for Project updating view"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

    def test_update_project(self):
        """Test Project updating with taskflow"""

        # Assert precondition
        self.assertEqual(Project.objects.all().count(), 2)

        request_data = model_to_dict(self.project)
        request_data['title'] = 'updated title'
        request_data['description'] = 'updated description'
        request_data['owner'] = self.user.sodar_uuid  # NOTE: Must add owner
        request_data['readme'] = 'updated readme'
        request_data['parent'] = str(self.category.sodar_uuid)
        request_data['public_guest_access'] = True
        request_data.update(
            app_settings.get_all_settings(project=self.project, post_safe=True)
        )  # Add default settings
        request_data['sodar_url'] = self.live_server_url  # HACK

        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:update',
                    kwargs={'project': self.project.sodar_uuid},
                ),
                request_data,
            )

        # Assert Project state after update
        self.assertEqual(Project.objects.all().count(), 2)
        self.project.refresh_from_db()

        expected = {
            'id': self.project.pk,
            'title': 'updated title',
            'type': PROJECT_TYPE_PROJECT,
            'parent': self.category.pk,
            'submit_status': SUBMIT_STATUS_OK,
            'description': 'updated description',
            'public_guest_access': True,
            'full_title': self.category.title + ' / updated title',
            'has_public_children': False,
            'sodar_uuid': self.project.sodar_uuid,
        }

        model_dict = model_to_dict(self.project)
        model_dict.pop('readme', None)
        self.assertEqual(model_dict, expected)
        self.assertEqual(self.project.readme.raw, 'updated readme')

        # Assert redirect
        with self.login(self.user):
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:detail',
                    kwargs={'project': self.project.sodar_uuid},
                ),
            )


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentCreateView(TestTaskflowBase):
    """Tests for RoleAssignment creation view"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        self.user_new = self.make_user('guest')

    def test_create_assignment(self):
        """Test RoleAssignment creation with taskflow"""
        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 2)

        # Issue POST request
        self.request_data.update(
            {
                'project': self.project.sodar_uuid,
                'user': self.user_new.sodar_uuid,
                'role': self.role_guest.pk,
            }
        )

        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:role_create',
                    kwargs={'project': self.project.sodar_uuid},
                ),
                self.request_data,
            )

        # Assert RoleAssignment state after creation
        self.assertEqual(RoleAssignment.objects.all().count(), 3)
        role_as = RoleAssignment.objects.get(
            project=self.project, user=self.user_new
        )
        self.assertIsNotNone(role_as)

        expected = {
            'id': role_as.pk,
            'project': self.project.pk,
            'user': self.user_new.pk,
            'role': self.role_guest.pk,
            'sodar_uuid': role_as.sodar_uuid,
        }

        self.assertEqual(model_to_dict(role_as), expected)

        # Assert redirect
        with self.login(self.user):
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:roles',
                    kwargs={'project': self.project.sodar_uuid},
                ),
            )


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentUpdateView(TestTaskflowBase):
    """Tests for RoleAssignment update view with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create guest user and role
        self.user_new = self.make_user('newuser')
        self.role_as = self._make_assignment_taskflow(
            self.project, self.user_new, self.role_guest
        )

    def test_update_assignment(self):
        """Test RoleAssignment updating with taskflow"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        self.request_data.update(
            {
                'project': self.project.sodar_uuid,
                'user': self.user_new.sodar_uuid,
                'role': self.role_contributor.pk,
            }
        )

        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:role_update',
                    kwargs={'roleassignment': self.role_as.sodar_uuid},
                ),
                self.request_data,
            )

        # Assert RoleAssignment state after update
        self.assertEqual(RoleAssignment.objects.all().count(), 3)
        role_as = RoleAssignment.objects.get(
            project=self.project, user=self.user_new
        )
        self.assertIsNotNone(role_as)

        expected = {
            'id': role_as.pk,
            'project': self.project.pk,
            'user': self.user_new.pk,
            'role': self.role_contributor.pk,
            'sodar_uuid': role_as.sodar_uuid,
        }

        self.assertEqual(model_to_dict(role_as), expected)

        # Assert redirect
        with self.login(self.user):
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:roles',
                    kwargs={'project': self.project.sodar_uuid},
                ),
            )


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentOwnerTransferView(TestTaskflowBase):
    """Tests for ownership transfer view with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create guest user and role
        self.user_new = self.make_user('newuser')
        self.role_as = self._make_assignment(
            self.project, self.user_new, self.role_guest
        )

    def test_transfer_owner(self):
        """Test ownership transfer with taskflow"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        self.request_data.update(
            {
                'project': self.project.sodar_uuid,
                'user': self.user_new.sodar_uuid,
                'role': self.role_contributor.pk,
            }
        )

        with self.login(self.user):
            self.client.post(
                reverse(
                    'projectroles:role_owner_transfer',
                    kwargs={'project': self.project.sodar_uuid},
                ),
                data={
                    'project': self.project.sodar_uuid,
                    'old_owner_role': self.role_guest.pk,
                    'new_owner': self.user_new.sodar_uuid,
                },
            )

        # Assert RoleAssignment state after update
        self.assertEqual(RoleAssignment.objects.all().count(), 3)
        role_as = RoleAssignment.objects.get(
            project=self.project, user=self.user_new
        )
        self.assertEqual(role_as.role, self.role_owner)
        # TODO: Test resulting users in iRODS once we do issue #387


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentDeleteView(TestTaskflowBase):
    """Tests for RoleAssignment delete view"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create guest user and role
        self.user_new = self.make_user('newuser')
        self.role_as = self._make_assignment_taskflow(
            self.project, self.user_new, self.role_guest
        )

    def test_delete_assignment(self):
        """Test RoleAssignment deleting with taskflow"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:role_delete',
                    kwargs={'roleassignment': self.role_as.sodar_uuid},
                ),
                self.request_data,
            )

        # Assert RoleAssignment state after update
        self.assertEqual(RoleAssignment.objects.all().count(), 2)

        # Assert redirect
        with self.login(self.user):
            self.assertRedirects(
                response,
                reverse(
                    'projectroles:roles',
                    kwargs={'project': self.project.sodar_uuid},
                ),
            )


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestProjectInviteAcceptView(ProjectInviteMixin, TestTaskflowBase):
    """Tests for ProjectInvite accepting view with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create guest user and role
        self.user_new = self.make_user('newuser')

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=True)
    @override_settings(AUTH_LDAP_USERNAME_DOMAIN='EXAMPLE')
    @override_settings(AUTH_LDAP_DOMAIN_PRINTABLE='EXAMPLE')
    @override_settings(ENABLE_LDAP=True)
    def test_accept_invite_ldap(self):
        """Test LDAP user accepting an invite with taskflow"""

        # Init invite
        invite = self._make_invite(
            email=INVITE_EMAIL,
            project=self.project,
            role=self.role_contributor,
            issuer=self.user,
            message='',
        )

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.filter(active=True).count(), 1)

        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project,
                user=self.user_new,
                role=self.role_contributor,
            ).count(),
            0,
        )

        with self.login(self.user_new):
            response = self.client.get(
                reverse(
                    'projectroles:invite_accept',
                    kwargs={'secret': invite.secret},
                ),
                self.request_data,
                follow=True,
            )

            self.assertListEqual(
                response.redirect_chain,
                [
                    (
                        reverse(
                            'projectroles:invite_process_ldap',
                            kwargs={'secret': invite.secret},
                        ),
                        302,
                    ),
                    (
                        reverse('home'),
                        302,
                    ),
                ],
            )

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=True)
    @override_settings(AUTH_LDAP_USERNAME_DOMAIN='EXAMPLE')
    @override_settings(AUTH_LDAP_DOMAIN_PRINTABLE='EXAMPLE')
    @override_settings(ENABLE_LDAP=True)
    def test_accept_invite_ldap_category(self):
        """Test LDAP user accepting an invite with taskflow for a category"""

        # Init invite
        invite = self._make_invite(
            email=INVITE_EMAIL,
            project=self.category,
            role=self.role_contributor,
            issuer=self.user,
            message='',
        )

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.filter(active=True).count(), 1)

        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.category,
                user=self.user_new,
                role=self.role_contributor,
            ).count(),
            0,
        )

        with self.login(self.user_new):
            response = self.client.get(
                reverse(
                    'projectroles:invite_accept',
                    kwargs={'secret': invite.secret},
                ),
                self.request_data,
                follow=True,
            )

            self.assertListEqual(
                response.redirect_chain,
                [
                    (
                        reverse(
                            'projectroles:invite_process_ldap',
                            kwargs={'secret': invite.secret},
                        ),
                        302,
                    ),
                    (
                        reverse(
                            'projectroles:detail',
                            kwargs={'project': self.category.sodar_uuid},
                        ),
                        302,
                    ),
                ],
            )

    def test_accept_invite_local(self):
        """Test local user accepting an invite with taskflow"""

        # Init invite
        invite = self._make_invite(
            email=INVITE_EMAIL,
            project=self.project,
            role=self.role_contributor,
            issuer=self.user,
            message='',
        )

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.filter(active=True).count(), 1)

        with self.login(self.user_new):
            response = self.client.get(
                reverse(
                    'projectroles:invite_accept',
                    kwargs={'secret': invite.secret},
                ),
                self.request_data,
                follow=True,
            )

            self.assertListEqual(
                response.redirect_chain,
                [
                    (
                        reverse(
                            'projectroles:invite_process_local',
                            kwargs={'secret': invite.secret},
                        ),
                        302,
                    ),
                    (
                        reverse('home'),
                        302,
                    ),
                ],
            )

            # Assert postconditions
            self.assertEqual(
                ProjectInvite.objects.filter(active=True).count(), 1
            )

    def test_accept_invite_local_category(self):
        """Test local user accepting an invite with taskflow for a category"""

        # Init invite
        invite = self._make_invite(
            email=INVITE_EMAIL,
            project=self.category,
            role=self.role_contributor,
            issuer=self.user,
            message='',
        )

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.filter(active=True).count(), 1)

        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.category,
                user=self.user_new,
                role=self.role_contributor,
            ).count(),
            0,
        )

        with self.login(self.user_new):
            response = self.client.get(
                reverse(
                    'projectroles:invite_accept',
                    kwargs={'secret': invite.secret},
                ),
                self.request_data,
                follow=True,
            )

            self.assertListEqual(
                response.redirect_chain,
                [
                    (
                        reverse(
                            'projectroles:invite_process_local',
                            kwargs={'secret': invite.secret},
                        ),
                        302,
                    ),
                    (
                        reverse('home'),
                        302,
                    ),
                ],
            )

            # Assert postconditions
            self.assertEqual(
                ProjectInvite.objects.filter(active=True).count(), 1
            )



"""Taskflow API views for the projectroles app"""

import json

from django.conf import settings

from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.views import APIView

from projectroles.models import Project, RoleAssignment, Role
from projectroles.views import SUBMIT_STATUS_OK, User, app_settings


class TaskflowAPIAuthentication(BaseAuthentication):
    """Taskflow API authentication handling"""

    def authenticate(self, request):
        taskflow_secret = None

        if request.method == 'POST' and 'sodar_secret' in request.POST:
            taskflow_secret = request.POST['sodar_secret']

        elif request.method == 'GET':
            taskflow_secret = request.GET.get('sodar_secret', None)

        if (
            not hasattr(settings, 'TASKFLOW_SODAR_SECRET')
            or taskflow_secret != settings.TASKFLOW_SODAR_SECRET
        ):
            raise PermissionDenied('Not authorized')


class TaskflowAPIPermission(BasePermission):
    """Taskflow API permission handling"""

    def has_permission(self, request, view):
        # Only allow accessing Taskflow API views if Taskflow is used
        return True if 'taskflow' in settings.ENABLED_BACKEND_PLUGINS else False


class BaseTaskflowAPIView(APIView):
    """Base Taskflow API view"""

    authentication_classes = [TaskflowAPIAuthentication]
    permission_classes = [TaskflowAPIPermission]


class TaskflowProjectGetAPIView(BaseTaskflowAPIView):
    """Taskflow API view for getting a project"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid'],
                submit_status=SUBMIT_STATUS_OK,
            )

        except Project.DoesNotExist as ex:
            return Response(str(ex), status=404)

        ret_data = {
            'project_uuid': str(project.sodar_uuid),
            'title': project.title,
            'description': project.description,
        }

        return Response(ret_data, status=200)


class TaskflowProjectUpdateAPIView(BaseTaskflowAPIView):
    """Taskflow API view for updating a project"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )
            parent = (
                Project.objects.get(sodar_uuid=request.data['parent_uuid'])
                if request.data.get('parent_uuid')
                else None
            )
            project.parent = parent
            project.title = request.data['title']
            project.description = request.data.get('description') or ''
            project.readme.raw = request.data['readme']
            project.public_guest_access = (
                request.data.get('public_guest_access') or False
            )
            project.save()

        except Project.DoesNotExist as ex:
            return Response(str(ex), status=404)

        return Response('ok', status=200)


class TaskflowRoleAssignmentGetAPIView(BaseTaskflowAPIView):
    """Taskflow API view for getting a role assignment for user and project"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )
            user = User.objects.get(sodar_uuid=request.data['user_uuid'])

        except (Project.DoesNotExist, User.DoesNotExist) as ex:
            return Response(str(ex), status=404)

        try:
            role_as = RoleAssignment.objects.get(project=project, user=user)
            ret_data = {
                'assignment_uuid': str(role_as.sodar_uuid),
                'project_uuid': str(role_as.project.sodar_uuid),
                'user_uuid': str(role_as.user.sodar_uuid),
                'role_pk': role_as.role.pk,
                'role_name': role_as.role.name,
            }
            return Response(ret_data, status=200)

        except RoleAssignment.DoesNotExist as ex:
            return Response(str(ex), status=404)


class TaskflowRoleAssignmentSetAPIView(BaseTaskflowAPIView):
    """Taskflow API view for creating or updating a role assignment"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )
            user = User.objects.get(sodar_uuid=request.data['user_uuid'])
            role = Role.objects.get(pk=request.data['role_pk'])

        except (
            Project.DoesNotExist,
            User.DoesNotExist,
            Role.DoesNotExist,
        ) as ex:
            return Response(str(ex), status=404)

        try:
            role_as = RoleAssignment.objects.get(project=project, user=user)
            role_as.role = role
            role_as.save()

        except RoleAssignment.DoesNotExist:
            role_as = RoleAssignment(project=project, user=user, role=role)
            role_as.save()

        return Response('ok', status=200)


class TaskflowRoleAssignmentDeleteAPIView(BaseTaskflowAPIView):
    """Taskflow API view for deleting a role assignment"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )
            user = User.objects.get(sodar_uuid=request.data['user_uuid'])

        except (Project.DoesNotExist, User.DoesNotExist) as ex:
            return Response(str(ex), status=404)

        try:
            role_as = RoleAssignment.objects.get(project=project, user=user)
            role_as.delete()

        except RoleAssignment.DoesNotExist as ex:
            return Response(str(ex), status=404)

        return Response('ok', status=200)


class TaskflowProjectSettingsGetAPIView(BaseTaskflowAPIView):
    """Taskflow API view for getting project settings"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )

        except Project.DoesNotExist as ex:
            return Response(str(ex), status=404)

        project_settings = app_settings.get_all_settings(project)

        for k, v in project_settings.items():
            if isinstance(v, dict):
                project_settings[k] = json.dumps(v)

        ret_data = {
            'project_uuid': project.sodar_uuid,
            'settings': project_settings,
        }

        return Response(ret_data, status=200)


class TaskflowProjectSettingsSetAPIView(BaseTaskflowAPIView):
    """Taskflow API view for updating project settings"""

    def post(self, request):
        try:
            project = Project.objects.get(
                sodar_uuid=request.data['project_uuid']
            )

        except Project.DoesNotExist as ex:
            return Response(str(ex), status=404)

        for k, v in json.loads(request.data['settings']).items():
            app_settings.set_app_setting(
                k.split('.')[1], k.split('.')[2], v, project
            )

        return Response('ok', status=200)



"""Timeline API for adding and updating events"""
import logging
import re

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils.text import Truncator

# Projectroles dependency
from projectroles.models import Project, RemoteSite
from projectroles.plugins import get_app_plugin
from projectroles.templatetags.projectroles_common_tags import get_user_html
from projectroles.utils import get_app_names

from timeline.models import (
    ProjectEvent,
    ProjectEventObjectRef,
    EVENT_STATUS_TYPES,
)


logger = logging.getLogger(__name__)
User = get_user_model()


# Local variables
APP_NAMES = get_app_names()
LABEL_MAX_WIDTH = 32
UNKNOWN_LABEL = '(unknown)'


class TimelineAPI:
    """Timeline backend API to be used by Django apps."""

    # Internal Helpers ---------------------------------------------------------

    @classmethod
    def _get_label(cls, label):
        """Format label to be displayed"""
        if not {' ', '-'}.intersection(label):
            return Truncator(label).chars(LABEL_MAX_WIDTH)
        return label

    @classmethod
    def _get_history_link(cls, ref_obj):
        """Return history link HTML for event object reference"""
        url_name = 'timeline:list_object_site'
        url_kwargs = {
            'object_model': ref_obj.object_model,
            'object_uuid': ref_obj.object_uuid,
        }
        if ref_obj.event.project:
            url_name = 'timeline:list_object'
            url_kwargs['project'] = ref_obj.event.project.sodar_uuid
        history_url = reverse(url_name, kwargs=url_kwargs)
        return (
            '<a href="{}" class="sodar-tl-object-link">'
            '<i class="iconify" '
            'data-icon="mdi:clock-time-eight-outline"></i></a>'.format(
                history_url
            )
        )

    @classmethod
    def _get_not_found_label(cls, ref_obj):
        """Get label for object which is not found in the database"""
        return '<span class="text-danger">{}</span> {}'.format(
            cls._get_label(ref_obj.name), cls._get_history_link(ref_obj)
        )

    @classmethod
    def _get_project_desc(cls, ref_obj, request=None):
        """Get description HTML for special case: Project model"""
        project = Project.objects.filter(sodar_uuid=ref_obj.object_uuid).first()
        if (
            project
            and request
            and request.user.has_perm('projectroles.view_project', project)
        ):
            return '<a class="sodar-tl-project-link" href="{}">{}</a>'.format(
                reverse(
                    'projectroles:detail',
                    kwargs={'project': project.sodar_uuid},
                ),
                cls._get_label(project.title),
            )
        elif project:
            return '<span class="text-danger">{}</span>'.format(
                cls._get_label(project.title)
            )
        return ref_obj.name

    @classmethod
    def _get_remote_site_desc(cls, ref_obj, request=None):
        """Get description HTML for special case: RemoteSite model"""
        site = RemoteSite.objects.filter(sodar_uuid=ref_obj.object_uuid).first()
        if site and request and request.user.is_superuser:
            return '<a href="{}">{}</a> {}'.format(
                reverse(
                    'projectroles:remote_projects',
                    kwargs={'remotesite': site.sodar_uuid},
                ),
                site.name,
                cls._get_history_link(ref_obj),
            )
        elif site:
            return site.name
        return cls._get_not_found_label(ref_obj)

    @classmethod
    def _get_ref_description(cls, event, ref_label, app_plugin, request):
        """
        Get reference object description for event description, or unknown label
        if not found.

        :param event: ProjectEvent object
        :param ref_label: Label for the reference object (string)
        :param app_plugin: App plugin or None
        :param request: Request object or None
        :return: String (contains HTML)
        """
        # Special case: Extra data reference
        if ref_label.startswith('extra-'):
            desc = app_plugin.get_extra_data_link(event.extra_data, ref_label)
            return desc if desc else UNKNOWN_LABEL

        # Get object reference
        try:
            ref_obj = ProjectEventObjectRef.objects.get(
                event=event, label=ref_label
            )
        except ProjectEventObjectRef.DoesNotExist:
            return UNKNOWN_LABEL

        # Special case: User model
        if ref_obj.object_model == 'User':
            try:
                user = User.objects.get(sodar_uuid=ref_obj.object_uuid)
                return '{} {}'.format(
                    get_user_html(user), cls._get_history_link(ref_obj)
                )
            except User.DoesNotExist:
                return UNKNOWN_LABEL

        # Special case: Project model
        elif ref_obj.object_model == 'Project':
            return cls._get_project_desc(ref_obj, request)

        # Special case: RemoteSite model
        elif ref_obj.object_model == 'RemoteSite':
            return cls._get_remote_site_desc(ref_obj, request)

        # Special case: projectroles app
        elif event.app == 'projectroles':
            return cls._get_not_found_label(ref_obj)

        # Apps with plugins
        else:
            try:
                link_data = app_plugin.get_object_link(
                    ref_obj.object_model, ref_obj.object_uuid
                )
            except Exception:
                link_data = None
            if link_data:
                return '<a href="{}" {}>{}</a> {}'.format(
                    link_data['url'],
                    (
                        'target="_blank"'
                        if 'blank' in link_data and link_data['blank'] is True
                        else ''
                    ),
                    cls._get_label(link_data['label']),
                    cls._get_history_link(ref_obj),
                )
            else:
                return cls._get_not_found_label(ref_obj)

    # API functions ------------------------------------------------------------

    @classmethod
    def add_event(
        cls,
        project,
        app_name,
        user,
        event_name,
        description,
        classified=False,
        extra_data=None,
        status_type=None,
        status_desc=None,
        status_extra_data=None,
    ):
        """
        Create and save a timeline event.

        :param project: Project object or None
        :param app_name: Name of app from which event was invoked (must
            correspond to "name" attribute of app plugin)
        :param user: User invoking the event or None
        :param event_name: Event ID string (must match schema)
        :param description: Description of status change (may include {object
            label} references)
        :param classified: Whether event is classified (boolean, optional)
        :param extra_data: Additional event data (dict, optional)
        :param status_type: Initial status type (string, optional)
        :param status_desc: Initial status description (string, optional)
        :param status_extra_data: Extra data for initial status (dict, optional)
        :return: ProjectEvent object
        :raise: ValueError if app_name or status_type is invalid
        """
        if app_name not in APP_NAMES:
            raise ValueError(
                'Unknown app name "{}" (active apps: {})'.format(
                    app_name, ', '.join(x for x in APP_NAMES)
                )
            )
        if status_type and status_type not in EVENT_STATUS_TYPES:
            raise ValueError(
                'Unknown status type "{}" (valid types: {})'.format(
                    status_type, ', '.join(x for x in EVENT_STATUS_TYPES)
                )
            )

        # Handle user in case called with AnonymousUser object
        if user and user.is_anonymous:
            user = None

        event = ProjectEvent()
        event.project = project
        event.app = app_name
        event.user = user
        event.event_name = event_name
        event.description = description
        event.classified = classified
        if extra_data:
            event.extra_data = extra_data
        event.save()

        # Always add "INIT" status when creating, except for "INFO"
        if status_type not in ['INFO', 'INIT']:
            event.set_status('INIT')
        # Add additional status if set (use if e.g. event is immediately "OK")
        if status_type:
            event.set_status(status_type, status_desc, status_extra_data)

        return event

    @classmethod
    def get_project_events(cls, project, classified=False):
        """
        Return timeline events for a project.

        :param project: Project object
        :param classified: Include classified (boolean)
        :return: QuerySet
        """
        events = ProjectEvent.objects.filter(project=project)
        if not classified:
            events = events.filter(classified=False)
        return events

    @classmethod
    def get_event_description(cls, event, plugin_lookup=None, request=None):
        """
        Return the description of a timeline event as HTML.

        :param event: ProjectEvent object
        :param plugin_lookup: App plugin lookup dict (optional)
        :param request: Request object (optional)
        :return: String (contains HTML)
        """
        desc = event.description
        ref_ids = re.findall('{\'?(.*?)\'?}', desc)
        if len(ref_ids) == 0:
            return event.description

        refs = {}
        app_plugin = None

        if event.app != 'projectroles':
            if plugin_lookup:
                app_plugin = plugin_lookup.get(event.app)
            else:
                app_plugin = get_app_plugin(event.app)
            if not app_plugin:
                msg = 'Plugin not found: {}'.format(event.app)
                logger.error(msg + ' (UUID={})'.format(event.sodar_uuid))
                return (
                    '<span class="sodar-tl-plugin-error text-danger">'
                    '{}</span>'.format(msg)
                )

        # Get links for object references
        for r in ref_ids:
            refs[r] = cls._get_ref_description(event, r, app_plugin, request)

        try:
            return event.description.format(**refs)
        except Exception as ex:  # Dispaly exception instead of crashing
            logger.error(
                'Error formatting event description: {} (UUID={})'.format(
                    ex, event.sodar_uuid
                )
            )
            return (
                '<span class="sodar-tl-format-error text-danger">'
                '{}: {}</span>'.format(ex.__class__.__name__, ex)
            )

    @classmethod
    def get_object_url(cls, obj, project=None):
        """
        Return the URL for a timeline event object history.

        :param obj: Django database object
        :param project: Related Project object or None
        :return: String
        """
        url_name = 'timeline:list_object_site'
        url_kwargs = {
            'object_model': obj.__class__.__name__,
            'object_uuid': obj.sodar_uuid,
        }
        if project:
            url_name = 'timeline:list_object'
            url_kwargs['project'] = project.sodar_uuid
        return reverse(url_name, kwargs=url_kwargs)

    @classmethod
    def get_object_link(cls, obj, project=None):
        """
        Return an inline HTML icon link for a timeline event object history.

        :param obj: Django database object
        :param project: Related Project object or None
        :return: String (contains HTML)
        """
        return (
            '<a href="{}" class="sodar-tl-object-link">'
            '<i class="iconify" data-icon="mdi:clock-time-eight-outline"></i>'
            '</a>'.format(cls.get_object_url(obj, project))
        )

    @classmethod
    def get_models(cls):
        """
        Return project event model classes for custom/advanced queries.

        :return: ProjectEvent, ProjectEventObjectRef
        """
        return ProjectEvent, ProjectEventObjectRef



"""REST API view permission tests for the filesfolders app"""

import os
import uuid

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions_api import (
    TestCoreProjectAPIPermissionBase,
)
from projectroles.utils import build_secret

from filesfolders.models import File, Folder, HyperLink
from filesfolders.tests.test_models import (
    FileMixin,
    FolderMixin,
    HyperLinkMixin,
)


# Local constants
SECRET = '7dqq83clo2iyhg29hifbor56og6911r5'
TEST_DATA_PATH = os.path.dirname(__file__) + '/data/'
ZIP_PATH_NO_FILES = TEST_DATA_PATH + 'no_files.zip'


class TestFolderAPIPermissions(FolderMixin, TestCoreProjectAPIPermissionBase):
    """Tests for Folder API view permissions"""

    def setUp(self):
        super().setUp()
        self.folder = self._make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,  # Project owner is the owner of folder
            description='',
        )

    def test_folder_list(self):
        """Test permissions for folder listing"""
        url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_list_anon(self):
        """Test permissions for folder listing with anonymous access"""
        url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_folder_create(self):
        """Test permissions for folder creation"""
        url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        request_data = {
            'name': 'New Folder',
            'flag': 'IMPORTANT',
            'description': 'Folder\'s description',
        }

        def _cleanup():
            Folder.objects.order_by('-pk').first().delete()

        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.guest_as.user, self.user_no_roles]
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=request_data
        )
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=request_data,
            cleanup_method=_cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_create_anon(self):
        """Test permissions for folder creation with anonymous access"""
        url = reverse(
            'filesfolders:api_folder_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        request_data = {
            'name': 'New Folder',
            'flag': 'IMPORTANT',
            'description': 'Folder\'s description',
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=request_data
        )

    def test_folder_retrieve(self):
        """Test permissions for folder retrieval"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200, method='GET')
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_retrieve_anon(self):
        """Test permissions for folder retrieval with anonymous access"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_folder_update_put(self):
        """Test permissions for folder updating with PUT"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        request_data = {
            'name': 'UPDATED Folder',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
        }
        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of folder
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=request_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PUT', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_update_put_anon(self):
        """Test permissions for folder updating with anonymous access"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        request_data = {
            'name': 'UPDATED Folder',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=request_data
        )

    def test_folder_update_patch(self):
        """Test permissions for folder updating with PATCH"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        request_data = {'name': 'UPDATED Folder'}
        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of folder
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, good_users, 200, method='PATCH', data=request_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PATCH', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_update_patch_anon(self):
        """Test permissions for folder updating with anonymous access"""
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': self.folder.sodar_uuid},
        )
        request_data = {'name': 'UPDATED Folder'}
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PATCH', data=request_data
        )

    def test_folder_destroy(self):
        """Test permissions for folder destroying with DELETE"""
        obj_uuid = uuid.uuid4()
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': obj_uuid},
        )

        def _make_folder():
            folder = self._make_folder(
                name='folder',
                project=self.project,
                folder=None,
                owner=self.owner_as.user,
                description='',
            )
            folder.sodar_uuid = obj_uuid
            folder.save()

        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of folder
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        _make_folder()
        self.assert_response_api(
            url, good_users, 204, method='DELETE', cleanup_method=_make_folder
        )
        self.assert_response_api(url, bad_users, 403, method='DELETE')
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            url,
            good_users,
            204,
            method='DELETE',
            cleanup_method=_make_folder,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 403, method='DELETE')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_destroy_anon(self):
        """Test permissions for folder destroying with anonymous access"""
        folder = self._make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,
            description='',
        )
        url = reverse(
            'filesfolders:api_folder_retrieve_update_destroy',
            kwargs={'folder': folder.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')


class TestFileAPIPermissions(FileMixin, TestCoreProjectAPIPermissionBase):
    """Tests for File API view permissions"""

    def setUp(self):
        super().setUp()
        self.file_content = bytes('content'.encode('utf-8'))
        self.file = self._make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.owner_as.user,
            description='',
            public_url=True,
            secret=SECRET,
        )
        self.new_file_name = 'New File'
        self.request_data = {
            'name': self.new_file_name,
            'flag': 'IMPORTANT',
            'description': 'File\'s description',
            'secret': 'foo',
            'public_url': True,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }

    def tearDown(self):
        self.request_data['file'].close()
        super().tearDown()

    def test_file_list(self):
        """Test permissions for file listing"""
        url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_list_anon(self):
        """Test permissions for file listing with anonymous access"""
        url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_file_create(self):
        """Test permissions for file creation"""
        url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )

        # NOTE: Must call this for ALL requests to seek the file
        def _cleanup():
            file = File.objects.filter(name=self.new_file_name).first()
            if file:
                file.delete()
            self.request_data['file'].seek(0)

        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.guest_as.user, self.user_no_roles]
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            bad_users,
            403,
            method='POST',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url,
            self.user_no_roles,
            403,
            method='POST',
            format='multipart',
            data=self.request_data,
        )
        self.request_data['file'].close()

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_create_anon(self):
        """Test permissions for file creation with anonymous access"""
        url = reverse(
            'filesfolders:api_file_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='POST',
            format='multipart',
            data=self.request_data,
        )
        self.request_data['file'].close()

    def test_file_retrieve(self):
        """Test permissions for file retrieval"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_retrieve_anon(self):
        """Test permissions for file retrieval with anonymous access"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_file_update_put(self):
        """Test permissions for file updating with PUT"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.request_data.update(
            {
                'name': 'UPDATED Folder',
                'flag': 'FLAG',
                'description': 'UPDATED Description',
            }
        )

        # NOTE: Must call this for ALL requests to seek the file
        def _cleanup():
            self.request_data['file'].seek(0)

        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of file
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url,
            good_users,
            200,
            method='PUT',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            bad_users,
            403,
            method='PUT',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            good_users,
            200,
            method='PUT',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url,
            self.user_no_roles,
            403,
            method='PUT',
            format='multipart',
            data=self.request_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_update_put_anon(self):
        """Test permissions for file updating with anonymous access"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.request_data.update(
            {
                'name': 'UPDATED Folder',
                'flag': 'FLAG',
                'description': 'UPDATED Description',
            }
        )
        self.project.set_public()
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='PUT',
            format='multipart',
            data=self.request_data,
        )

    def test_file_update_patch(self):
        """Test permissions for file updating with PATCH"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.request_data.update({'name': 'UPDATED Folder'})

        # NOTE: Must call this for ALL requests to seek the file
        def _cleanup():
            self.request_data['file'].seek(0)

        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of file
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url,
            good_users,
            200,
            method='PATCH',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            bad_users,
            403,
            method='PATCH',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            good_users,
            200,
            method='PATCH',
            format='multipart',
            data=self.request_data,
            cleanup_method=_cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url,
            self.user_no_roles,
            403,
            method='PATCH',
            format='multipart',
            data=self.request_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_update_patch_anon(self):
        """Test permissions for file updating with anonymous access"""
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.request_data.update({'name': 'UPDATED Folder'})
        self.project.set_public()
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='PATCH',
            format='multipart',
            data=self.request_data,
        )

    def test_file_destroy(self):
        """Test permissions for file destroying with DELETE"""
        obj_uuid = uuid.uuid4()
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': obj_uuid},
        )

        def _make_file():
            file = self._make_file(
                name='file2.txt',
                file_name='file2.txt',
                file_content=self.file_content,
                project=self.project,
                folder=None,
                owner=self.owner_as.user,
                description='',
                public_url=True,
                secret=build_secret(),
            )
            file.sodar_uuid = obj_uuid
            file.save()

        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of file
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        _make_file()
        self.assert_response_api(
            url, good_users, 204, method='DELETE', cleanup_method=_make_file
        )
        self.assert_response_api(url, bad_users, 403, method='DELETE')
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            url,
            good_users,
            204,
            method='DELETE',
            cleanup_method=_make_file,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url,
            self.user_no_roles,
            403,
            method='DELETE',
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_destroy_anon(self):
        """Test permissions for file destroying with anonymous access"""
        file = self._make_file(
            name='file2.txt',
            file_name='file2.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.owner_as.user,
            description='',
            public_url=True,
            secret=build_secret(),
        )
        url = reverse(
            'filesfolders:api_file_retrieve_update_destroy',
            kwargs={'file': file.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='DELETE',
        )

    def test_file_serve(self):
        """Test permissions for file serving"""
        url = reverse(
            'filesfolders:api_file_serve', kwargs={'file': self.file.sodar_uuid}
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200, method='GET')
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_serve_anon(self):
        """Test permissions for file serving with anonymous access"""
        url = reverse(
            'filesfolders:api_file_serve', kwargs={'file': self.file.sodar_uuid}
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)


class TestHyperLinkAPIPermissions(
    HyperLinkMixin, TestCoreProjectAPIPermissionBase
):
    """Tests for HyperLink API view permissions"""

    def setUp(self):
        super().setUp()
        self.hyperlink = self._make_hyperlink(
            name='Link',
            url='http://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,
            description='',
        )

    def test_hyperlink_list(self):
        """Test permissions for hyperlink listing"""
        url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_list_anon(self):
        """Test permissions for hyperlink listing with anonymous access"""
        url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_hyperlink_create(self):
        """Test permissions for hyperlink creation"""
        url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        request_data = {
            'name': 'New HyperLink',
            'flag': 'IMPORTANT',
            'description': 'HyperLink\'s description',
            'url': 'http://www.cubi.bihealth.org',
        }

        def _cleanup():
            HyperLink.objects.order_by('-pk').first().delete()

        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.guest_as.user, self.user_no_roles]
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=request_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=request_data
        )
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=request_data,
            cleanup_method=_cleanup,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_create_anon(self):
        """Test permissions for hyperlink creation with anonymous access"""
        url = reverse(
            'filesfolders:api_hyperlink_list_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        request_data = {
            'name': 'New HyperLink',
            'flag': 'IMPORTANT',
            'description': 'HyperLink\'s description',
            'url': 'http://www.cubi.bihealth.org',
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=request_data
        )

    def test_hyperlink_retrieve(self):
        """Test permissions for hyperlink retrieval"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200, method='GET')
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_retrieve_anon(self):
        """Test permissions for hyperlink retrieval with anonymous access"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_hyperlink_update_put(self):
        """Test permissions for hyperlink updating with PUT"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        request_data = {
            'name': 'UPDATED HyperLink',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
            'url': 'http://www.bihealth.org',
        }
        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of link
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=request_data
        )
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=request_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PUT', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_update_put_anon(self):
        """Test permissions for hyperlink updating with anonymous access"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        request_data = {
            'name': 'UPDATED HyperLink',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
            'url': 'http://www.bihealth.org',
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=request_data
        )

    def test_hyperlink_update_patch(self):
        """Test permissions for hyperlink updating with PATCH"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        request_data = {'name': 'UPDATED Hyperlink'}
        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of link
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PATCH', data=request_data
        )
        self.assert_response_api(
            url, good_users, 200, method='PATCH', data=request_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PATCH', data=request_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_update_patch_anon(self):
        """Test permissions for hyperlink updating with anonymous access"""
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': self.hyperlink.sodar_uuid},
        )
        request_data = {'name': 'UPDATED Hyperlink'}
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PATCH', data=request_data
        )

    def test_hyperlink_destroy(self):
        """Test permissions for hyperlink destroying with DELETE"""
        obj_uuid = uuid.uuid4()
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': obj_uuid},
        )

        def _make_hyperlink():
            link = self._make_hyperlink(
                name='New Link',
                url='http://www.duckduckgo.com/',
                project=self.project,
                folder=None,
                owner=self.owner_as.user,
                description='',
            )
            link.sodar_uuid = obj_uuid
            link.save()

        good_users = [
            self.superuser,
            self.owner_as.user,  # Owner of link
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        _make_hyperlink()
        self.assert_response_api(
            url,
            good_users,
            204,
            method='DELETE',
            cleanup_method=_make_hyperlink,
        )
        self.assert_response_api(url, bad_users, 403, method='DELETE')
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')
        self.assert_response_api(
            url,
            good_users,
            204,
            method='DELETE',
            cleanup_method=_make_hyperlink,
            knox=True,
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url,
            self.user_no_roles,
            403,
            method='DELETE',
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_destroy_anon(self):
        """Test permissions for hyperlink destroying with anonymous access"""
        link = self._make_hyperlink(
            name='New Link',
            url='http://www.duckduckgo.com/',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,
            description='',
        )
        url = reverse(
            'filesfolders:api_hyperlink_retrieve_update_destroy',
            kwargs={'hyperlink': link.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='DELETE',
        )



# Projectroles dependency
from projectroles.plugins import BackendPluginPoint

from .api import TaskflowAPI


class BackendPlugin(BackendPluginPoint):
    """Plugin for registering backend app with Projectroles"""

    #: Name (slug-safe, used in URLs)
    name = 'taskflow'

    #: Title (used in templates)
    title = 'Taskflow'

    #: Iconify icon
    icon = 'mdi:database'

    #: Description string
    description = 'SODAR Taskflow backend for data transactions'

    def get_api(self):
        """Return API entry point object."""
        return TaskflowAPI()



"""View tests for the tokens app"""


from django.urls import reverse

from knox.models import AuthToken

from test_plus.test import TestCase


class TestUserTokenListView(TestCase):
    """Tests for UserTokenListView"""

    def setUp(self):
        self.user = self.make_user()

    def _make_token(self):
        self.tokens = [AuthToken.objects.create(self.user, None)]

    def test_list_empty(self):
        """Test that rendering the list view without any tokens works"""
        with self.login(self.user):
            response = self.get('tokens:list')
        self.response_200(response)
        self.assertEqual(len(response.context['object_list']), 0)

    def test_list_one(self):
        """Test that rendering the list view with one token works"""
        self._make_token()
        with self.login(self.user):
            response = self.get('tokens:list')
        self.response_200(response)
        self.assertEqual(len(response.context['object_list']), 1)


class TestUserTokenCreateView(TestCase):
    """Tests for UserTokenCreateView"""

    def setUp(self):
        self.user = self.make_user()

    def test_get(self):
        """Test that showing the creation form works"""
        with self.login(self.user):
            response = self.get('tokens:create')
        self.response_200(response)
        self.assertIsNotNone(response.context["form"])

    def test_post_success_no_ttl(self):
        """Test creating an authentication token with TTL=0 works"""
        self.assertEqual(AuthToken.objects.count(), 0)
        with self.login(self.user):
            response = self.post('tokens:create', data={'ttl': 0})
        self.response_200(response)
        self.assertEqual(AuthToken.objects.count(), 1)

    def test_post_success_with_ttl(self):
        """Test creating an authentication token with TTL != 0 works"""
        self.assertEqual(AuthToken.objects.count(), 0)
        with self.login(self.user):
            response = self.post('tokens:create', data={'ttl': 10})
        self.response_200(response)
        self.assertEqual(AuthToken.objects.count(), 1)


class TestUserTokenDeleteView(TestCase):
    """Tests for UserTokenDeleteView"""

    def setUp(self):
        self.user = self.make_user()
        AuthToken.objects.create(user=self.user)
        self.token = AuthToken.objects.first()

    def test_get(self):
        """Test that showing the deletion form works"""
        with self.login(self.user):
            response = self.get('tokens:delete', pk=self.token.pk)
        self.response_200(response)

    def test_post_success(self):
        """Test that deleting a token works"""
        self.assertEqual(AuthToken.objects.count(), 1)
        with self.login(self.user):
            response = self.post('tokens:delete', pk=self.token.pk)
        self.response_302(response)
        self.assertRedirects(
            response, reverse('tokens:list'), fetch_redirect_response=False
        )



"""REST API view permission tests for the projectroles app"""

import uuid

from django.test import override_settings
from django.urls import reverse

from projectroles.models import (
    Project,
    RoleAssignment,
    ProjectInvite,
    SODAR_CONSTANTS,
)
from projectroles.tests.test_permissions import TestProjectPermissionBase
from projectroles.tests.test_views_api import SODARAPIViewTestMixin
from projectroles.views_api import CORE_API_MEDIA_TYPE, CORE_API_DEFAULT_VERSION

from rest_framework.test import APITestCase

NEW_PROJECT_TITLE = 'New Project'


# Base Classes and Mixins ------------------------------------------------------


class SODARAPIPermissionTestMixin(SODARAPIViewTestMixin):
    """Mixin for permission testing with knox auth"""

    def assert_response_api(
        self,
        url,
        users,
        status_code,
        method='GET',
        format='json',
        data=None,
        media_type=None,
        version=None,
        knox=False,
        cleanup_method=None,
        req_kwargs=None,
    ):
        """
        Assert a response status code for url with API headers and optional
        Knox token authentication. Creates a Knox token for each user where
        needed.

        :param url: Target URL for the request
        :param users: Users to test (single user, list or tuple)
        :param status_code: Status code
        :param method: Method for request (default="GET")
        :param format: Request format (string, default="json")
        :param data: Optional data for request (dict)
        :param media_type: String (default = cls.media_type)
        :param version: String (default = cls.api_version)
        :param knox: Use Knox token auth instead of Django login (boolean)
        :param cleanup_method: Callable method to clean up data after a
               successful request
        :param req_kwargs: Optional request kwargs override (dict or None)
        """
        if cleanup_method and not callable(cleanup_method):
            raise ValueError('cleanup_method is not callable')

        def _send_request():
            req_method = getattr(self.client, method.lower(), None)
            if not req_method:
                raise ValueError('Invalid method "{}"'.format(method))
            if req_kwargs:  # Override request kwargs if set
                r_kwargs.update(req_kwargs)
            return req_method(url, **r_kwargs)

        if not isinstance(users, (list, tuple)):
            users = [users]

        for user in users:
            r_kwargs = {'format': format}
            if data:
                r_kwargs['data'] = data
            if knox and not user:  # Anonymous
                raise ValueError(
                    'Unable to test Knox token auth with anonymous user'
                )
            r_kwargs.update(self.get_accept_header(media_type, version))

            if knox:
                r_kwargs.update(self.get_token_header(self.get_token(user)))
                response = _send_request()
            elif user:
                with self.login(user):
                    response = _send_request()
            else:  # Anonymous, no knox
                response = _send_request()

            msg = 'user={}; content="{}"'.format(user, response.content)
            self.assertEqual(response.status_code, status_code, msg=msg)

            if cleanup_method:
                cleanup_method()


class TestProjectAPIPermissionBase(
    SODARAPIPermissionTestMixin, APITestCase, TestProjectPermissionBase
):
    """Base class for testing project permissions in SODAR API views"""


class TestCoreProjectAPIPermissionBase(
    SODARAPIPermissionTestMixin, APITestCase, TestProjectPermissionBase
):
    """
    Base class for testing project permissions in internal SODAR Core API views
    """

    media_type = CORE_API_MEDIA_TYPE
    api_version = CORE_API_DEFAULT_VERSION


# Tests ------------------------------------------------------------------------


class TestAPIPermissions(TestCoreProjectAPIPermissionBase):
    """Tests for projectroles API view permissions"""

    def test_project_list(self):
        """Test permissions for ProjectListAPIView"""
        url = reverse('projectroles:api_project_list')
        good_users = [
            self.superuser,
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)

    def test_project_retrieve(self):
        """Test permissions for ProjectRetrieveAPIView"""
        url = reverse(
            'projectroles:api_project_retrieve',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        self.assert_response_api(url, bad_users, 403, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_retrieve_anon(self):
        """Test permissions for ProjectRetrieveAPIView with anonymous access"""
        url = reverse(
            'projectroles:api_project_retrieve',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 200)

    def test_project_create_root(self):
        """Test permissions for ProjectCreateAPIView with no parent"""
        url = reverse('projectroles:api_project_create')
        post_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY'],
            'parent': '',
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }
        good_users = [self.superuser]
        bad_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]

        def _cleanup():
            p = Project.objects.filter(title=NEW_PROJECT_TITLE).first()
            if p:
                p.delete()

        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_create_root_anon(self):
        """Test permissions for ProjectCreateAPIView with no parent with anonymous access"""
        url = reverse('projectroles:api_project_create')
        post_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY'],
            'parent': '',
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )

    def test_project_create(self):
        """Test permissions for ProjectCreateAPIView"""
        url = reverse('projectroles:api_project_create')
        post_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'],
            'parent': str(self.category.sodar_uuid),
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }

        def _cleanup():
            p = Project.objects.filter(title=NEW_PROJECT_TITLE).first()
            if p:
                p.delete()

        good_users = [self.superuser, self.owner_as_cat.user]
        bad_users = [
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_create_anon(self):
        """Test permissions for ProjectCreateAPIView with anonymous access"""
        url = reverse('projectroles:api_project_create')
        post_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'],
            'parent': str(self.category.sodar_uuid),
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )

    def test_project_update(self):
        """Test permissions for ProjectUpdateAPIView"""
        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.project.sodar_uuid},
        )
        put_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'],
            'parent': str(self.category.sodar_uuid),
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=put_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=put_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=put_data
        )
        # Test with Knox
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=put_data, knox=True
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=put_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PUT', data=put_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_update_anon(self):
        """Test permissions for ProjectUpdateAPIView with anonymous access"""
        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.project.sodar_uuid},
        )
        put_data = {
            'title': NEW_PROJECT_TITLE,
            'type': SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'],
            'parent': str(self.category.sodar_uuid),
            'description': 'description',
            'readme': 'readme',
            'owner': str(self.owner_as.user.sodar_uuid),
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=put_data
        )

    def test_role_create(self):
        """Test permissions for RoleAssignmentCreateAPIView"""
        # Create user for assignments
        assign_user = self.make_user('assign_user')
        url = reverse(
            'projectroles:api_role_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
            'user': str(assign_user.sodar_uuid),
        }
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]

        def _cleanup():
            role_as = RoleAssignment.objects.filter(
                project=self.project,
                role__name=SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
                user=assign_user,
            ).first()
            if role_as:
                role_as.delete()

        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_role_create_anon(self):
        """Test permissions for RoleAssignmentCreateAPIView with anonymous access"""
        assign_user = self.make_user('assign_user')
        url = reverse(
            'projectroles:api_role_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
            'user': str(assign_user.sodar_uuid),
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )

    def test_role_update(self):
        """Test permissions for RoleAssignmentUpdateAPIView"""
        # Create user and assignment
        assign_user = self.make_user('assign_user')
        update_as = self._make_assignment(
            self.project, assign_user, self.role_contributor
        )
        url = reverse(
            'projectroles:api_role_update',
            kwargs={'roleassignment': update_as.sodar_uuid},
        )
        put_data = {
            'role': self.role_guest.name,
            'user': str(assign_user.sodar_uuid),
        }
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=put_data
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=put_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=put_data
        )
        # Test with Knox
        self.assert_response_api(
            url, good_users, 200, method='PUT', data=put_data, knox=True
        )
        self.assert_response_api(
            url, bad_users, 403, method='PUT', data=put_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='PUT', data=put_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_role_update_anon(self):
        """Test permissions for RoleAssignmentUpdateAPIView with anonymous access"""
        # Create user and assignment
        assign_user = self.make_user('assign_user')
        update_as = self._make_assignment(
            self.project, assign_user, self.role_contributor
        )
        url = reverse(
            'projectroles:api_role_update',
            kwargs={'roleassignment': update_as.sodar_uuid},
        )
        put_data = {
            'role': self.role_guest.name,
            'user': str(assign_user.sodar_uuid),
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='PUT', data=put_data
        )

    def test_role_delete(self):
        """Test permissions for RoleAssignmentDestroyAPIView"""
        # Create user and assignment
        assign_user = self.make_user('assign_user')
        role_uuid = uuid.uuid4()  # Ensure fixed uuid

        def _cleanup():
            update_as = self._make_assignment(
                self.project, assign_user, self.role_contributor
            )
            update_as.sodar_uuid = role_uuid
            update_as.save()

        url = reverse(
            'projectroles:api_role_destroy',
            kwargs={'roleassignment': role_uuid},
        )
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        _cleanup()
        self.assert_response_api(
            url, good_users, 204, method='DELETE', cleanup_method=_cleanup
        )
        self.assert_response_api(url, bad_users, 403, method='DELETE')
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            204,
            method='DELETE',
            cleanup_method=_cleanup,
            knox=True,
        )
        self.assert_response_api(
            url, bad_users, 403, method='DELETE', knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 403, method='DELETE')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_role_delete_anon(self):
        """Test permissions for RoleAssignmentDestroyAPIView with anonymous access"""
        # Create user and assignment
        assign_user = self.make_user('assign_user')
        role_uuid = uuid.uuid4()  # Ensure fixed uuid
        update_as = self._make_assignment(
            self.project, assign_user, self.role_contributor
        )
        update_as.sodar_uuid = role_uuid
        update_as.save()
        url = reverse(
            'projectroles:api_role_destroy',
            kwargs={'roleassignment': role_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 401, method='DELETE')

    def test_owner_transfer(self):
        """Test permissions for RoleAssignmentOwnerTransferAPIView"""
        # Create user for assignments
        self.new_owner = self.make_user('new_owner')
        self.new_owner_as = self._make_assignment(
            self.project, self.new_owner, self.role_contributor
        )
        url = reverse(
            'projectroles:api_role_owner_transfer',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'new_owner': self.new_owner.username,
            'old_owner_role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
        }

        def _cleanup():
            self.new_owner_as.refresh_from_db()
            self.new_owner_as.role = self.role_contributor
            self.new_owner_as.save()

            self.owner_as.refresh_from_db()
            self.owner_as.role = self.role_owner
            self.owner_as.save()

        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
        ]
        bad_users = [
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
            data=post_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
            data=post_data,
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_owner_transfer_anon(self):
        """Test permissions for RoleAssignmentOwnerTransferAPIView with anonymous access"""
        # Create user for assignments
        self.new_owner = self.make_user('new_owner')
        self.new_owner_as = self._make_assignment(
            self.project, self.new_owner, self.role_contributor
        )
        url = reverse(
            'projectroles:api_role_owner_transfer',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'new_owner': self.new_owner.username,
            'old_owner_role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )

    def test_invite_list(self):
        """Test permissions for ProjectInviteListAPIView"""
        url = reverse(
            'projectroles:api_invite_list',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, bad_users, 403)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_invite_list_anon(self):
        """Test permissions for ProjectInviteListAPIView with anonymous access"""
        url = reverse(
            'projectroles:api_invite_list',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 401)

    def test_invite_create(self):
        """Test permissions for ProjectInviteCreateAPIView"""
        email = 'new@example.com'
        url = reverse(
            'projectroles:api_invite_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'email': email,
            'role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
        }
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]

        def _cleanup():
            invite = ProjectInvite.objects.filter(
                email=email,
            ).first()
            if invite:
                invite.delete()

        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data
        )
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            201,
            method='POST',
            data=post_data,
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url, bad_users, 403, method='POST', data=post_data, knox=True
        )
        # Test public project
        self.project.set_public()
        self.assert_response_api(
            url, self.user_no_roles, 403, method='POST', data=post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_invite_create_anon(self):
        """Test permissions for ProjectInviteCreateAPIView with anonymous access"""
        email = 'new@example.com'
        url = reverse(
            'projectroles:api_invite_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'email': email,
            'role': SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR'],
        }
        self.project.set_public()
        self.assert_response_api(
            url, self.anonymous, 401, method='POST', data=post_data
        )

    def test_invite_revoke(self):
        """Test permissions for ProjectInviteRevokeAPIView"""
        self.invite = self._make_invite(
            email='new@example.com',
            project=self.project,
            role=self.role_contributor,
            issuer=self.user_owner,
        )
        url = reverse(
            'projectroles:api_invite_revoke',
            kwargs={'projectinvite': self.invite.sodar_uuid},
        )
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]

        def _cleanup():
            self.invite.active = True
            self.invite.save()

        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
            cleanup_method=_cleanup,
        )
        self.assert_response_api(
            url,
            bad_users,
            403,
            method='POST',
        )
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='POST',
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
            knox=True,
            cleanup_method=_cleanup,
        )
        self.assert_response_api(url, bad_users, 403, method='POST', knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 403, method='POST')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_invite_revoke_anon(self):
        """Test permissions for ProjectInviteRevokeAPIView with anonymous access"""
        self.invite = self._make_invite(
            email='new@example.com',
            project=self.project,
            role=self.role_contributor,
            issuer=self.user_owner,
        )
        url = reverse(
            'projectroles:api_invite_revoke',
            kwargs={'projectinvite': self.invite.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 401, method='POST')

    def test_invite_resend(self):
        """Test permissions for ProjectInviteResendAPIView"""
        self.invite = self._make_invite(
            email='new@example.com',
            project=self.project,
            role=self.role_contributor,
            issuer=self.user_owner,
        )
        url = reverse(
            'projectroles:api_invite_resend',
            kwargs={'projectinvite': self.invite.sodar_uuid},
        )
        good_users = [
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
        ]
        bad_users = [
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
        )
        self.assert_response_api(
            url,
            bad_users,
            403,
            method='POST',
        )
        self.assert_response_api(
            url,
            self.anonymous,
            401,
            method='POST',
        )
        # Test with Knox
        self.assert_response_api(
            url,
            good_users,
            200,
            method='POST',
            knox=True,
        )
        self.assert_response_api(url, bad_users, 403, method='POST', knox=True)
        # Test public project
        self.project.set_public()
        self.assert_response_api(url, self.user_no_roles, 403, method='POST')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_invite_resend_anon(self):
        """Test permissions for ProjectInviteResendAPIView with anonymous access"""
        self.invite = self._make_invite(
            email='new@example.com',
            project=self.project,
            role=self.role_contributor,
            issuer=self.user_owner,
        )
        url = reverse(
            'projectroles:api_invite_resend',
            kwargs={'projectinvite': self.invite.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response_api(url, self.anonymous, 401, method='POST')

    def test_user_list(self):
        """Test permissions for UserListAPIView"""
        url = reverse('projectroles:api_user_list')
        good_users = [
            self.superuser,
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)

    def test_user_current(self):
        """Test permissions for CurrentUserRetrieveAPIView"""
        url = reverse('projectroles:api_user_current')
        good_users = [
            self.superuser,
            self.owner_as_cat.user,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response_api(url, good_users, 200)
        self.assert_response_api(url, self.anonymous, 401)
        self.assert_response_api(url, good_users, 200, knox=True)



"""Permission tests for the tokens app"""

from django.test import override_settings
from django.urls import reverse

from knox.models import AuthToken

# Projectroles dependency
from projectroles.tests.test_permissions import TestPermissionBase


class TestTokenPermissions(TestPermissionBase):
    """Tests for token view permissions"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        # No user
        self.anonymous = None

    def test_list(self):
        """Test permissions for token list"""
        url = reverse('tokens:list')
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_list_anon(self):
        """Test permissions for token list with anonymous access"""
        url = reverse('tokens:list')
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    def test_create(self):
        """Test permissions for token creation"""
        url = reverse('tokens:create')
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_create_anon(self):
        """Test permissions for token creation with anonymous access"""
        url = reverse('tokens:create')
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    def test_delete(self):
        """Test permissions for token deletion"""
        token = AuthToken.objects.create(self.regular_user, None)
        url = reverse('tokens:delete', kwargs={'pk': token[0].pk})
        good_users = [self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)



"""Ajax API views for the timeline app"""

from django.utils.timezone import localtime

from rest_framework.response import Response

# Projectroles dependency
from projectroles.views_ajax import (
    SODARBaseProjectAjaxView,
    SODARBasePermissionAjaxView,
)

from timeline.models import ProjectEvent
from timeline.templatetags.timeline_tags import get_status_style


class EventDetailMixin:
    """Mixin for event detail retrieval helpers"""

    def get_event_details(self, event):
        """
        Return event details.

        :param event: ProjectEvent object
        :return: Dict
        """
        ret = {
            'app': event.app,
            'name': event.event_name,
            'user': event.user.username if event.user else 'N/A',
            'timestamp': localtime(event.get_timestamp()).strftime(
                '%Y-%m-%d %H:%M:%S'
            ),
            'status': [],
        }
        status_changes = event.get_status_changes(reverse=True)
        for s in status_changes:
            ret['status'].append(
                {
                    'timestamp': localtime(s.timestamp).strftime(
                        '%Y-%m-%d %H:%M:%S'
                    ),
                    'description': s.description,
                    'type': s.status_type,
                    'class': get_status_style(s),
                }
            )
        return ret


class ProjectEventDetailAjaxView(EventDetailMixin, SODARBaseProjectAjaxView):
    """Ajax view for retrieving event details for projects"""

    permission_required = 'timeline.view_timeline'
    allow_anonymous = True

    def get(self, request, *args, **kwargs):
        event = ProjectEvent.objects.filter(
            sodar_uuid=self.kwargs['projectevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_event', event.project
        ):
            return Response(status=403)
        return Response(self.get_event_details(event), status=200)


class SiteEventDetailAjaxView(EventDetailMixin, SODARBasePermissionAjaxView):
    """Ajax view for retrieving event details for site-wide events"""

    permission_required = 'timeline.view_site_timeline'
    allow_anonymous = True

    def get(self, request, *args, **kwargs):
        event = ProjectEvent.objects.filter(
            sodar_uuid=self.kwargs['projectevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_site_event'
        ):
            return Response(status=403)
        return Response(self.get_event_details(event), status=200)



"""Factory Boy factory classes for ``bgjobs``."""

import factory
from bgjobs.models import BackgroundJob


class BackgroundJobFactory(factory.django.DjangoModelFactory):
    """Factory for ``BackgroundJobFactory`` model."""

    class Meta:
        model = BackgroundJob

    project = None  # Can't set this because of circular dependency. factory.SubFactory(ProjectFactory)
    user = None  # Wait for SODAR core to offer a UserFactory
    job_type = 'type'



from django.conf.urls import url

from . import views

app_name = 'userprofile'

urlpatterns = [
    url(regex=r'^profile$', view=views.UserDetailView.as_view(), name='detail'),
    url(
        regex=r'^profile/settings/update$',
        view=views.UserSettingUpdateView.as_view(),
        name='settings_update',
    ),
]
